==> tinyinit-0.1.tbz: tinyinit-0.1/begetty.c <==
/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/wait.h>

#include "init.h"

#if	defined(__gnu_linux__)
static const char binlogin[] = "/bin/login";
#elif	defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
static const char binlogin[] = "/usr/bin/login";
#else
#error	"Unknown operating system"
#endif

static const struct option long_options[] =
{
	{	"login",	1,	0,	'l'	},
	{	NULL,		0,	NULL,	0	}
};

static int termsig;
static int chldsig;
static void handler(int sig)
{
	if (sig == SIGTERM)
		termsig = 1;
	else	/* (sig == SIGCHLD) */
		chldsig = 1;
}

static int getattr(const char *dev, struct termios *tp)
{
	int fd;
	int ret;

	fd = open(dev, (O_RDWR|O_NOCTTY));
	if (fd == -1)
	{
		perror(dev);
		return (-1);
	}
	ret = tcgetattr(fd, tp);
	if (ret == -1)
		perror(dev);

	(void) close(fd);
	return (ret);
}

static int setattr(const char *dev, struct termios *tp)
{
	int fd;
	int ret;

	fd = open(dev, (O_RDWR|O_NOCTTY));
	if (fd == -1)
	{
		perror(dev);
		return (-1);
	}
	ret = tcsetattr(fd, TCSAFLUSH, tp);
	if (ret == -1)
		perror(dev);

	(void) close(fd);
	return (ret);
}

#define	FATALERROR	111		/* Fatal error exit code */

int main(int argc, char *argv[])
{
	const char *login;
	int c;
	char dev[64];
	sigset_t mask;
	struct sigaction sa;
	struct termios termios;

	login = binlogin;

	opterr = 0;
	while ((c = getopt_long(argc, argv, "+", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 'l':
			login = optarg;
			if (strlen(login) == 0)	/* i.e. "--login=" */
				goto optdone;
			break;
		}
		if (c == '?')
			break;
	}
optdone:

	if (c != -1 || (optind + 1) != argc)
	{
		fprintf(stderr, "usage:\t%s [--login=pathname] port\n", argv[0]);
		return (1);
	}

	if (strlen(argv[optind]) > (sizeof(dev)-5-1))
	{
		fprintf(stderr, "Port path exceeds %zu characters\n", (sizeof(dev)-1));
		return (1);
	}
	strcpy(dev, "/dev/");
	strcat(dev, argv[optind]);

	/* Block TERM/CHLD signals */
	sigemptyset(&mask);
	sigaddset(&mask, SIGTERM);
	sigaddset(&mask, SIGCHLD);
	if (sigprocmask(SIG_SETMASK, &mask, &mask) == -1)
		return (1);

	/* Catch TERM/CHLD signals */
	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGTERM, &sa, NULL) == -1)
		return (1);
	if (sigaction(SIGCHLD, &sa, NULL) == -1)
		return (1);

	if (getattr(dev, &termios) == -1)
		return (1);

	for ( ; ; )
	{
		pid_t pid;
		int status;

		pid = fork();
		if (pid == -1)
		{
			perror("fork");
			return (1);
		}

		if (pid == 0)
		{
			/* Child */

			if (setsid() == -1)
				return (FATALERROR);

			if (set_stdio(dev) == -1)
				return (FATALERROR);

			if (ioctl(STDIN_FILENO, TIOCSCTTY, 0) == -1)
				return (FATALERROR);

			(void) sigprocmask(SIG_SETMASK, &mask, NULL);

			(void) execl(login, login, (char *)NULL);
			return (FATALERROR);
		}

		/* Parent */

		/* Wait for TERM/CHLD signals */
		(void) sigsuspend(&mask);

		if (termsig)
		{
			termsig = 0;	/* reset flag */
			(void) kill_wait(pid, SIGHUP);
			return (0);
		}

		if (chldsig)
		{
			chldsig = 0;	/* reset flag */
			if (waitpid(pid, &status, 0) == -1)
			{
				perror("waitpid");
				return (1);
			}
			if (WIFEXITED(status) && WEXITSTATUS(status) == FATALERROR)
			{
				fprintf(stderr, "%s: Fatal error\n", login);
				return (1);
			}
		}

		if (setattr(dev, &termios) == -1)
			return (1);
	}
}

==> tinyinit-0.3.tbz: tinyinit-0.3/begetty.c <==
/*
 *	Copyright 2011 Paul Parsons
 *
 *	This file is part of TinyInit.
 *
 *	TinyInit is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.
 *
 *	TinyInit is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with TinyInit.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <fcntl.h>
#include <getopt.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/wait.h>

#include "init.h"

#if	defined(__gnu_linux__)
static const char binlogin[] = "/bin/login";
#elif	defined(__FreeBSD__) || defined(__NetBSD__) || defined(__OpenBSD__)
static const char binlogin[] = "/usr/bin/login";
#else
#error	"Unknown operating system"
#endif

static const struct option long_options[] =
{
	{	"login",	1,	0,	'l'	},
	{	NULL,		0,	NULL,	0	}
};

static int termsig;
static int chldsig;
static void handler(int sig)
{
	if (sig == SIGTERM)
		termsig = 1;
	else	/* (sig == SIGCHLD) */
		chldsig = 1;
}

static int getattr(const char *dev, struct termios *tp)
{
	int fd;
	int ret;

	fd = open(dev, (O_RDWR|O_NOCTTY));
	if (fd == -1)
	{
		perror(dev);
		return (-1);
	}
	ret = tcgetattr(fd, tp);
	if (ret == -1)
		perror(dev);

	(void) close(fd);
	return (ret);
}

static int setattr(const char *dev, struct termios *tp)
{
	int fd;
	int ret;

	fd = open(dev, (O_RDWR|O_NOCTTY));
	if (fd == -1)
	{
		perror(dev);
		return (-1);
	}
	ret = tcsetattr(fd, TCSAFLUSH, tp);
	if (ret == -1)
		perror(dev);

	(void) close(fd);
	return (ret);
}

#define	FATALERROR	111		/* Fatal error exit code */

int main(int argc, char *argv[])
{
	const char *login;
	int c;
	char dev[64];
	sigset_t mask;
	struct sigaction sa;
	struct termios termios;

	login = binlogin;

	opterr = 0;
	while ((c = getopt_long(argc, argv, "+", long_options, NULL)) != -1)
	{
		switch (c)
		{
		case 'l':
			login = optarg;
			if (strlen(login) == 0)	/* i.e. "--login=" */
				goto optdone;
			break;
		}
		if (c == '?')
			break;
	}
optdone:

	if (c != -1 || (optind + 1) != argc)
	{
		fprintf(stderr, "usage:\tbegetty [--login=pathname] port\n");
		return (1);
	}

	if (strlen(argv[optind]) > (sizeof(dev)-5-1))
	{
		fprintf(stderr, "Port path exceeds %zu characters\n", (sizeof(dev)-1));
		return (1);
	}
	strcpy(dev, "/dev/");
	strcat(dev, argv[optind]);

	/* Block TERM/CHLD signals */
	sigemptyset(&mask);
	sigaddset(&mask, SIGTERM);
	sigaddset(&mask, SIGCHLD);
	if (sigprocmask(SIG_SETMASK, &mask, &mask) == -1)
		return (1);

	/* Catch TERM/CHLD signals */
	sa.sa_handler = handler;
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_NOCLDSTOP;
	if (sigaction(SIGTERM, &sa, NULL) == -1)
		return (1);
	if (sigaction(SIGCHLD, &sa, NULL) == -1)
		return (1);

	if (getattr(dev, &termios) == -1)
		return (1);

	for ( ; ; )
	{
		pid_t pid;
		int status;

		pid = fork();
		if (pid == -1)
		{
			perror("fork");
			return (1);
		}

		if (pid == 0)
		{
			/* Child */

			if (setsid() == -1)
				return (FATALERROR);

			if (set_stdio(dev) == -1)
				return (FATALERROR);

			if (ioctl(STDIN_FILENO, TIOCSCTTY, 0) == -1)
				return (FATALERROR);

			(void) sigprocmask(SIG_SETMASK, &mask, NULL);

			(void) execl(login, login, (char *)NULL);
			return (FATALERROR);
		}

		/* Parent */

		/* Wait for TERM/CHLD signals */
		(void) sigsuspend(&mask);

		if (termsig)
		{
			termsig = 0;	/* reset flag */
			(void) kill_wait(pid, SIGHUP);
			return (0);
		}

		if (chldsig)
		{
			chldsig = 0;	/* reset flag */
			if (waitpid(pid, &status, 0) == -1)
			{
				perror("waitpid");
				return (1);
			}
			if (WIFEXITED(status) && WEXITSTATUS(status) == FATALERROR)
			{
				fprintf(stderr, "%s: Fatal error\n", login);
				return (1);
			}
		}

		if (setattr(dev, &termios) == -1)
			return (1);
	}
}
